
from __future__ import annotations

import hashlib
import json
from datetime import datetime
from typing import Optional, Sequence, List, Dict, Tuple, FrozenSet, \
    Iterable, Awaitable

from aioredis import Redis, MultiExecError  # type: ignore

from pymap.bytes import HashStream
from pymap.exceptions import MailboxAbort, MailboxNotFound, MailboxConflict
from pymap.flags import FlagOp
from pymap.interfaces.message import CachedMessage
from pymap.listtree import ListTree
from pymap.mailbox import MailboxSnapshot
from pymap.mime import MessageContent
from pymap.parsing.message import PreparedMessage
from pymap.parsing.modutf7 import modutf7_encode, modutf7_decode
from pymap.parsing.specials import ObjectId, FetchRequirement, SequenceSet
from pymap.parsing.specials.flag import Flag, Deleted, Seen
from pymap.selected import SelectedSet, SelectedMailbox
from pymap.threads import ThreadKey

from ._util import unwatch_pipe, watch_pipe, check_errors
from .cleanup import Cleanup
from .keys import NamespaceKeys, ContentKeys, MailboxKeys, MessageKeys
from .message import Message
from ..mailbox import SavedMessage, MailboxDataInterface, MailboxSetInterface

__all__ = ['Message', 'MailboxData', 'MailboxSet']


class MailboxData(MailboxDataInterface[Message]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxDataInterface`
    for the redis backend.

    """

    def __init__(self, redis: Redis, mailbox_id: bytes, uid_validity: int,
                 keys: MailboxKeys, ns_keys: NamespaceKeys,
                 cleanup: Cleanup) -> None:
        super().__init__()
        self._redis = redis
        self._mailbox_id = ObjectId(mailbox_id)
        self._uid_validity = uid_validity
        self._selected_set = SelectedSet()
        self._keys = keys
        self._ns_keys = ns_keys
        self._cleanup = cleanup

    @property
    def mailbox_id(self) -> ObjectId:
        return self._mailbox_id

    @property
    def readonly(self) -> bool:
        return False

    @property
    def uid_validity(self) -> int:
        return self._uid_validity

    @property
    def selected_set(self) -> SelectedSet:
        return self._selected_set

    async def update_selected(self, selected: SelectedMailbox) \
            -> SelectedMailbox:
        last_mod_seq = selected.mod_sequence
        if last_mod_seq is None:
            mod_seq, updated, expunged = await self._get_initial()
        else:
            mod_seq, updated, expunged = await self._get_updated(last_mod_seq)
        selected.mod_sequence = mod_seq
        selected.add_updates(updated, expunged)
        return selected

    async def save(self, message: bytes) -> SavedMessage:
        redis = self._redis
        ns_keys = self._ns_keys
        content = MessageContent.parse(message)
        new_email_id = ObjectId.random_email_id()
        msg_hash = HashStream(hashlib.sha1()).digest(content)
        thread_keys = ThreadKey.get_all(content.header)
        thread_key_keys = [b'\0'.join(thread_key)
                           for thread_key in thread_keys]
        await redis.unwatch()
        multi = redis.multi_exec()
        multi.hsetnx(ns_keys.email_ids, msg_hash, new_email_id.value)
        multi.hget(ns_keys.email_ids, msg_hash)
        if thread_key_keys:
            multi.hmget(ns_keys.thread_ids, *thread_key_keys)
        else:
            multi.hmget(ns_keys.thread_ids, b'')
        _, email_id, thread_ids = await multi.execute()
        thread_id_b = next((thread_id for thread_id in thread_ids
                            if thread_id is not None), None)
        if thread_id_b is None:
            thread_id = ObjectId.random_thread_id()
        else:
            thread_id = ObjectId(thread_id_b)
        ct_keys = ContentKeys(ns_keys, email_id)
        multi = redis.multi_exec()
        multi.hset(ct_keys.data, b'full', message)
        multi.hset(ct_keys.data, b'full-json',
                   json.dumps(content.json))
        multi.hset(ct_keys.data, b'header', bytes(content.header))
        multi.hset(ct_keys.data, b'header-json',
                   json.dumps(content.header.json))
        multi.expire(ct_keys.data, self._cleanup.content_expire)
        for thread_key_key in thread_key_keys:
            multi.hsetnx(ns_keys.thread_ids, thread_key_key, thread_id.value)
        await multi.execute()
        return SavedMessage(ObjectId(email_id), thread_id, None)

    async def add(self, message: PreparedMessage, *,
                  recent: bool = False) -> Message:
        redis = self._redis
        keys = self._keys
        ns_keys = self._ns_keys
        when = message.when or datetime.now()
        is_deleted = Deleted in message.flag_set
        is_unseen = Seen not in message.flag_set
        msg_flags = [flag.value for flag in message.flag_set]
        msg_time = when.isoformat().encode('ascii')
        email_id = message.email_id
        thread_id = message.thread_id
        ct_keys = ContentKeys(ns_keys, email_id)
        while True:
            pipe = watch_pipe(redis, keys.max_mod, keys.abort)
            pipe.mget(keys.max_uid, keys.max_mod, keys.abort)
            _, _, (max_uid, max_mod, abort) = await pipe.execute()
            MailboxAbort.assertFalse(abort)
            new_uid = int(max_uid or 0) + 1
            new_mod = int(max_mod or 0) + 1
            msg_keys = MessageKeys(keys, new_uid)
            multi = redis.multi_exec()
            multi.set(keys.max_uid, new_uid)
            multi.set(keys.max_mod, new_mod)
            multi.sadd(keys.uids, new_uid)
            multi.zadd(keys.mod_seq, new_mod, new_uid)
            multi.zadd(keys.seq, new_uid, new_uid)
            if recent:
                multi.sadd(keys.recent, new_uid)
            if is_deleted:
                multi.sadd(keys.deleted, new_uid)
            if is_unseen:
                multi.zadd(keys.unseen, new_uid, new_uid)
            if msg_flags:
                multi.sadd(msg_keys.flags, *msg_flags)
            multi.hmset(msg_keys.immutable, b'time', msg_time,
                        b'emailid', email_id.value,
                        b'threadid', thread_id.value)
            multi.persist(ct_keys.data)
            multi.hincrby(ns_keys.content_refs, email_id.value, 1)
            try:
                await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                break
        return Message(new_uid, when, message.flag_set,
                       email_id=email_id, thread_id=thread_id,
                       redis=redis, ns_keys=ns_keys)

    async def get(self, uid: int, cached_msg: CachedMessage = None,
                  requirement: FetchRequirement = FetchRequirement.METADATA) \
            -> Optional[Message]:
        redis = self._redis
        keys = self._keys
        ns_keys = self._ns_keys
        msg_keys = MessageKeys(keys, uid)
        await redis.unwatch()
        multi = redis.multi_exec()
        multi.sismember(keys.uids, uid)
        multi.smembers(msg_keys.flags)
        multi.hmget(msg_keys.immutable, b'time', b'emailid', b'threadid')
        multi.get(keys.abort)
        exists, flags, (time, email_id, thread_id), abort = \
            await multi.execute()
        MailboxAbort.assertFalse(abort)
        if not exists:
            if cached_msg is not None:
                if not isinstance(cached_msg, Message):
                    raise TypeError(cached_msg)
                return Message.copy_expunged(cached_msg)
            else:
                return None
        msg_flags = {Flag(flag) for flag in flags}
        msg_email_id = ObjectId.maybe(email_id)
        msg_thread_id = ObjectId.maybe(thread_id)
        msg_time = datetime.fromisoformat(time.decode('ascii'))
        return Message(uid, msg_time, msg_flags,
                       email_id=msg_email_id, thread_id=msg_thread_id,
                       redis=redis, ns_keys=ns_keys)

    async def delete(self, uids: Iterable[int]) -> None:
        redis = self._redis
        keys = self._keys
        uids = list(uids)
        if not uids:
            return
        while True:
            pipe = watch_pipe(redis, keys.max_mod, keys.abort)
            pipe.mget(keys.max_mod, keys.abort)
            _, _, (max_mod, abort) = await pipe.execute()
            MailboxAbort.assertFalse(abort)
            new_mod = int(max_mod or 0) + 1
            multi = redis.multi_exec()
            multi.set(keys.max_mod, new_mod)
            multi.srem(keys.uids, *uids)
            multi.zrem(keys.seq, *uids)
            multi.zrem(keys.mod_seq, *uids)
            multi.srem(keys.recent, *uids)
            multi.srem(keys.deleted, *uids)
            multi.zrem(keys.unseen, *uids)
            for uid in uids:
                multi.zadd(keys.expunged, new_mod, uid)
                msg_keys = MessageKeys(keys, uid)
                self._cleanup.add_message(multi, msg_keys)
            try:
                await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                break

    async def claim_recent(self, selected: SelectedMailbox) -> None:
        redis = self._redis
        keys = self._keys
        while True:
            pipe = watch_pipe(redis, keys.max_mod, keys.abort)
            pipe.smembers(keys.recent)
            pipe.mget(keys.max_mod, keys.abort)
            _, _, recent, (max_mod, abort) = await pipe.execute()
            if not recent:
                break
            MailboxAbort.assertFalse(abort)
            new_mod = int(max_mod or 0) + 1
            multi = redis.multi_exec()
            multi.set(keys.max_mod, new_mod)
            for uid in recent:
                multi.zadd(keys.mod_seq, new_mod, uid)
            multi.delete(keys.recent)
            try:
                await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                break
        for uid_bytes in recent:
            selected.session_flags.add_recent(int(uid_bytes))

    async def update_flags(self, messages: Sequence[Message],
                           flag_set: FrozenSet[Flag], mode: FlagOp) -> None:
        redis = self._redis
        keys = self._keys
        messages = list(messages)
        if not messages:
            return
        uids = {msg.uid: msg for msg in messages}
        while True:
            pipe = watch_pipe(redis, keys.max_mod, keys.abort)
            pipe.smembers(keys.uids)
            pipe.mget(keys.max_mod, keys.abort)
            _, _, existing_uids, (max_mod, abort) = await pipe.execute()
            MailboxAbort.assertFalse(abort)
            update_uids = uids.keys() & {int(uid) for uid in existing_uids}
            if not update_uids:
                break
            new_mod = int(max_mod or 0) + 1
            new_flags: Dict[int, Awaitable[Sequence[bytes]]] = {}
            multi = redis.multi_exec()
            multi.set(keys.max_mod, new_mod)
            for msg in messages:
                msg_uid = msg.uid
                if msg_uid not in update_uids:
                    continue
                msg_keys = MessageKeys(keys, msg_uid)
                flag_vals = (flag.value for flag in flag_set)
                multi.zadd(keys.mod_seq, new_mod, msg_uid)
                if mode == FlagOp.REPLACE:
                    multi.delete(msg_keys.flags)
                    if flag_set:
                        multi.sadd(msg_keys.flags, *flag_vals)
                    if Deleted in flag_set:
                        multi.sadd(keys.deleted, msg_uid)
                    else:
                        multi.srem(keys.deleted, msg_uid)
                    if Seen not in flag_set:
                        multi.zadd(keys.unseen, msg_uid, msg_uid)
                    else:
                        multi.zrem(keys.unseen, msg_uid)
                elif mode == FlagOp.ADD and flag_set:
                    multi.sadd(msg_keys.flags, *flag_vals)
                    if Deleted in flag_set:
                        multi.sadd(keys.deleted, msg_uid)
                    if Seen in flag_set:
                        multi.zrem(keys.unseen, msg_uid)
                elif mode == FlagOp.DELETE and flag_set:
                    multi.srem(msg_keys.flags, *flag_vals)
                    if Deleted in flag_set:
                        multi.srem(keys.deleted, msg_uid)
                    if Seen in flag_set:
                        multi.zadd(keys.unseen, msg_uid, msg_uid)
                new_flags[msg_uid] = multi.smembers(msg_keys.flags)
            try:
                await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                for msg_uid, msg_flags in new_flags.items():
                    msg = uids[msg_uid]
                    msg.permanent_flags = frozenset(
                        Flag(flag) for flag in await msg_flags)
                break

    async def find_deleted(self, seq_set: SequenceSet,
                           selected: SelectedMailbox) -> Sequence[int]:
        if not seq_set.is_all:
            return await super().find_deleted(seq_set, selected)
        redis = self._redis
        pipe = unwatch_pipe(redis)
        pipe.smembers(self._keys.deleted)
        _, deleted = await pipe.execute()
        return [int(uid) for uid in deleted]

    async def cleanup(self) -> None:
        pass

    async def snapshot(self) -> MailboxSnapshot:
        redis = self._redis
        keys = self._keys
        while True:
            pipe = watch_pipe(redis, keys.seq, keys.abort)
            pipe.mget(keys.max_uid, keys.abort)
            pipe.zcard(keys.seq)
            pipe.scard(keys.recent)
            pipe.zcard(keys.unseen)
            pipe.zrange(keys.unseen, 0, 0)
            _, _, (max_uid, abort), exists, num_recent, num_unseen, unseen = \
                await pipe.execute()
            MailboxAbort.assertFalse(abort)
            next_uid = int(max_uid or 0) + 1
            if not unseen:
                first_unseen: Optional[int] = None
                break
            first_uid = int(unseen[0])
            multi = redis.multi_exec()
            multi.zrank(keys.seq, first_uid)
            try:
                first_unseen_b, = await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                first_unseen = int(first_unseen_b) + 1
                break
        return MailboxSnapshot(self.mailbox_id, self.readonly,
                               self.uid_validity, self.permanent_flags,
                               self.session_flags, exists, num_recent,
                               num_unseen, first_unseen, next_uid)

    async def _get_initial(self) \
            -> Tuple[int, Sequence[Message], Sequence[int]]:
        redis = self._redis
        keys = self._keys
        ns_keys = self._ns_keys
        while True:
            pipe = watch_pipe(redis, keys.max_mod, keys.abort)
            pipe.zrange(keys.seq)
            pipe.get(keys.abort)
            _, _, uids, abort = await pipe.execute()
            MailboxAbort.assertFalse(abort)
            multi = redis.multi_exec()
            multi.get(keys.max_mod)
            for uid in uids:
                msg_keys = MessageKeys(keys, uid)
                multi.echo(uid)
                multi.smembers(msg_keys.flags)
                multi.hmget(msg_keys.immutable,
                            b'time', b'emailid', b'threadid')
            try:
                results = await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                break
        mod_seq = int(results[0] or 0)
        updated: List[Message] = []
        for i in range(1, len(results), 3):
            msg_uid = int(results[i])
            msg_flags = {Flag(flag) for flag in results[i + 1]}
            time_b, email_id, thread_id = results[i + 2]
            msg_time = datetime.fromisoformat(time_b.decode('ascii'))
            msg = Message(msg_uid, msg_time, msg_flags,
                          email_id=ObjectId(email_id),
                          thread_id=ObjectId(thread_id),
                          redis=redis, ns_keys=ns_keys)
            updated.append(msg)
        return mod_seq, updated, []

    async def _get_updated(self, last_mod_seq: int) \
            -> Tuple[int, Sequence[Message], Sequence[int]]:
        redis = self._redis
        keys = self._keys
        ns_keys = self._ns_keys
        while True:
            pipe = watch_pipe(redis, keys.max_mod, keys.abort)
            pipe.zrangebyscore(keys.mod_seq, last_mod_seq)
            pipe.get(keys.abort)
            _, _, uids, abort = await pipe.execute()
            MailboxAbort.assertFalse(abort)
            multi = redis.multi_exec()
            multi.get(keys.max_mod)
            multi.zrangebyscore(keys.expunged, last_mod_seq)
            for uid in uids:
                msg_keys = MessageKeys(keys, uid)
                multi.echo(uid)
                multi.smembers(msg_keys.flags)
                multi.hmget(msg_keys.immutable,
                            b'time', b'emailid', b'threadid')
            try:
                results = await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                break
        mod_seq = int(results[0] or 0)
        expunged = [int(uid) for uid in results[1]]
        updated: List[Message] = []
        for i in range(2, len(results), 3):
            msg_uid = int(results[i])
            msg_flags = {Flag(flag) for flag in results[i + 1]}
            time_b, email_id, thread_id = results[i + 2]
            msg_time = datetime.fromisoformat(time_b.decode('ascii'))
            msg = Message(msg_uid, msg_time, msg_flags,
                          email_id=ObjectId(email_id),
                          thread_id=ObjectId(thread_id),
                          redis=redis, ns_keys=ns_keys)
            updated.append(msg)
        return mod_seq, updated, expunged


class MailboxSet(MailboxSetInterface[MailboxData]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxSetInterface`
    for the redis backend.

    """

    def __init__(self, redis: Redis, keys: NamespaceKeys,
                 cleanup: Cleanup) -> None:
        super().__init__()
        self._redis = redis
        self._keys = keys
        self._cleanup = cleanup

    @property
    def delimiter(self) -> str:
        return '/'

    async def set_subscribed(self, name: str, subscribed: bool) -> None:
        pipe = unwatch_pipe(self._redis)
        if subscribed:
            pipe.sadd(self._keys.subscribed, name)
        else:
            pipe.srem(self._keys.subscribed, name)
        await pipe.execute()

    async def list_subscribed(self) -> ListTree:
        pipe = unwatch_pipe(self._redis)
        pipe.smembers(self._keys.subscribed)
        _, mailboxes_raw = await pipe.execute()
        mailboxes = [modutf7_decode(name) for name in mailboxes_raw]
        return ListTree(self.delimiter).update('INBOX', *mailboxes)

    async def list_mailboxes(self) -> ListTree:
        redis = self._redis
        await redis.unwatch()
        multi = redis.multi_exec()
        multi.hgetall(self._keys.mailboxes)
        multi.zrange(self._keys.order)
        mailboxes, mbx_order = await multi.execute()
        rev_mbx = {mbx_id: key for key, mbx_id in mailboxes.items()}
        mailboxes = [modutf7_decode(rev_mbx[mbx_id]) for mbx_id in mbx_order
                     if mbx_id in rev_mbx]
        return ListTree(self.delimiter).update('INBOX', *mailboxes)

    async def get_mailbox(self, name: str,
                          try_create: bool = False) -> MailboxData:
        redis = self._redis
        name_key = modutf7_encode(name)
        while True:
            pipe = watch_pipe(redis, self._keys.mailboxes)
            pipe.hget(self._keys.mailboxes, name_key)
            _, _, mbx_id = await pipe.execute()
            if mbx_id is None:
                raise MailboxNotFound(name, try_create)
            multi = redis.multi_exec()
            multi.hget(self._keys.uid_validity, name_key)
            try:
                uidval, *_ = await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                mbx_keys = MailboxKeys(self._keys, mbx_id)
                return MailboxData(redis, mbx_id, int(uidval), mbx_keys,
                                   self._keys, self._cleanup)

    async def add_mailbox(self, name: str) -> ObjectId:
        redis = self._redis
        name_key = modutf7_encode(name)
        while True:
            mbx_id = ObjectId.random_mailbox_id()
            pipe = watch_pipe(redis, self._keys.mailboxes)
            pipe.incr(self._keys.max_order)
            pipe.hexists(self._keys.mailboxes, name_key)
            _, _, order, exists = await pipe.execute()
            if exists:
                raise MailboxConflict(name)
            multi = redis.multi_exec()
            multi.hset(self._keys.mailboxes, name_key, mbx_id.value)
            multi.zadd(self._keys.order, order, mbx_id.value)
            multi.hincrby(self._keys.uid_validity, name_key)
            try:
                _, _, uidval = await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                break
        return mbx_id

    async def delete_mailbox(self, name: str) -> None:
        redis = self._redis
        name_key = modutf7_encode(name)
        await redis.unwatch()
        multi = redis.multi_exec()
        multi.hget(self._keys.mailboxes, name_key)
        multi.hdel(self._keys.mailboxes, name_key)
        mbx_id, _ = await multi.execute()
        if mbx_id is None:
            raise MailboxNotFound(name)
        mbx_keys = MailboxKeys(self._keys, mbx_id)
        pipe = redis.pipeline()
        pipe.zrem(self._keys.order, mbx_id)
        pipe.set(mbx_keys.abort, 1)
        self._cleanup.add_mailbox(pipe, mbx_keys)
        await pipe.execute()

    async def rename_mailbox(self, before: str, after: str) -> None:
        redis = self._redis
        while True:
            pipe = watch_pipe(redis, self._keys.mailboxes)
            pipe.hgetall(self._keys.mailboxes)
            _, _, all_keys = await pipe.execute()
            all_mbx = {modutf7_decode(key): ns for key, ns in all_keys.items()}
            tree = ListTree(self.delimiter).update('INBOX', *all_mbx.keys())
            before_entry = tree.get(before)
            after_entry = tree.get(after)
            if before_entry is None:
                raise MailboxNotFound(before)
            elif after_entry is not None:
                raise MailboxConflict(after)
            multi = redis.multi_exec()
            for before_name, after_name in tree.get_renames(before, after):
                before_id = all_mbx[before_name]
                before_key = modutf7_encode(before_name)
                after_key = modutf7_encode(after_name)
                multi.hset(self._keys.mailboxes, after_key, before_id)
                multi.hdel(self._keys.mailboxes, before_key)
                multi.hincrby(self._keys.uid_validity, after_key)
            if before == 'INBOX':
                inbox_id = ObjectId.random_mailbox_id()
                multi.hset(self._keys.mailboxes, b'INBOX', inbox_id.value)
                multi.hincrby(self._keys.uid_validity, b'INBOX')
            try:
                await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                break




from __future__ import annotations

import asyncio

from aioredis import Redis, WatchVariableError  # type: ignore

__all__ = ['unwatch_pipe', 'watch_pipe', 'check_errors']


def unwatch_pipe(redis: Redis) -> Redis:
    pipe = redis.pipeline()
    pipe.unwatch()
    return pipe


def watch_pipe(redis: Redis, *watch: bytes) -> Redis:
    pipe = unwatch_pipe(redis)
    pipe.watch(*watch)
    return pipe


async def check_errors(multi: Redis) -> bool:
    # Prevents warning about exception never being retrieved.
    errors = await asyncio.gather(*multi._results, return_exceptions=True)
    return any(not isinstance(exc, WatchVariableError) for exc in errors)




from __future__ import annotations

from typing import Type, Tuple
from typing_extensions import Final

from . import ParsingExpectation

__all__ = ['ExpectContinuation']


class ExpectContinuation(ParsingExpectation[memoryview]):
    """Indicates that the buffer has been successfully parsed so far, but
    requires a continuation of the command from the client.

    Args:
        message: The message from the server.
        literal_length: If the continuation is for a string literal, this
            is the byte length to expect.

    """

    __slots__ = ['message', 'literal_length']

    def __init__(self, message: bytes, literal_length: int = 0) -> None:
        super().__init__()
        self.message: Final = message
        self.literal_length: Final = literal_length

    @property
    def consume_args(self) -> Tuple[str, Type[memoryview]]:
        return 'continuations', memoryview




from __future__ import annotations

from typing import Optional, Type, Tuple, Dict, List, Collection

from . import Space, Params
from .command import Command
from .command.any import CapabilityCommand, LogoutCommand, NoOpCommand
from .command.auth import AppendCommand, CreateCommand, DeleteCommand, \
    ExamineCommand, ListCommand, LSubCommand, RenameCommand, SelectCommand, \
    StatusCommand, SubscribeCommand, UnsubscribeCommand
from .command.nonauth import AuthenticateCommand, LoginCommand, StartTLSCommand
from .command.select import CheckCommand, CloseCommand, ExpungeCommand, \
    CopyCommand, FetchCommand, StoreCommand, SearchCommand, UidCommand, \
    UidCopyCommand, UidExpungeCommand, UidFetchCommand, UidSearchCommand, \
    UidStoreCommand, IdleCommand
from .exceptions import NotParseable
from .primitives import Atom
from .specials import Tag

__all__ = ['builtin_commands', 'InvalidCommand', 'Commands']

#: List of built-in commands. These commands are automatically registered by a
#: new :class:`Commands` object.
builtin_commands: Collection[Type[Command]] = [
    CapabilityCommand, LogoutCommand, NoOpCommand, AppendCommand,
    CreateCommand, DeleteCommand, ExamineCommand, ListCommand, LSubCommand,
    RenameCommand, SelectCommand, StatusCommand, SubscribeCommand,
    UnsubscribeCommand, AuthenticateCommand, LoginCommand, StartTLSCommand,
    CheckCommand, CloseCommand, ExpungeCommand, CopyCommand, FetchCommand,
    StoreCommand, SearchCommand, UidCommand, UidCopyCommand, UidExpungeCommand,
    UidFetchCommand, UidSearchCommand, UidStoreCommand, IdleCommand]


class InvalidCommand(Command):
    """An invalid command, either because the tag or command name could not be
    parsed, the command name was not recognized, or a known command was given
    invalid arguments.

    Args:
        params: The parsing parameters
        parse_exc: The exception that caused the failure, if any.
        command: The command name, if available.
        command_type: The command type, if available.

    """

    command = b'[INVALID]'

    def __init__(self, params: Params, parse_exc: Optional[NotParseable],
                 command: Optional[bytes] = None,
                 command_type: Optional[Type[Command]] = None) -> None:
        super().__init__(params.tag)
        self._parse_exc = parse_exc
        self._command = command
        self._command_type = command_type

    @property
    def value(self) -> bytes:
        return self._command or b''

    @property
    def message(self) -> bytes:
        """The message to include in a BAD response."""
        if not self.command_name:
            return b'Command not given.'
        elif not self.command_type:
            return b'%b: Command not implemented.' % (self.command_name, )
        else:
            return b'%b: Invalid arguments.' % (self.command_name, )

    @property
    def command_name(self) -> Optional[bytes]:
        """The command name, if the name could be parsed."""
        return self._command

    @property
    def command_type(self) -> Optional[Type[Command]]:
        """The command type, if parsing failed due to invalid arguments."""
        return self._command_type

    @property
    def parse_exc(self) -> Optional[NotParseable]:
        """The parsing exception, if any."""
        return self._parse_exc

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[InvalidCommand, memoryview]:
        raise RuntimeError('Do not call')


class Commands:
    """Manages the set of all known IMAP commands and the ability to parse."""

    __slots__ = ['commands']

    def __init__(self) -> None:
        super().__init__()
        self.commands: Dict[bytes, Type[Command]] = {}
        self._load_commands()

    def _load_commands(self):
        for cmd in builtin_commands:
            self.register(cmd)

    def register(self, cmd: Type[Command]) -> None:
        """Register a new IMAP command.

        Args:
            cmd: The new command type.

        """
        self.commands[cmd.command] = cmd

    def deregister(self, name: bytes) -> None:
        """Deregister an IMAP command by name. Does nothing if the command is
        not registered.

        Args:
            name: The IMAP command name.

        """
        self.commands.pop(name.upper(), None)

    def parse(self, buf: memoryview, params: Params) \
            -> Tuple[Command, memoryview]:
        """Parse the given bytes into a command. The basic syntax is a tag
        string, a command name, possibly some arguments, and then an endline.
        If the command has a complete structure but cannot be parsed, an
        :class:`InvalidCommand` is returned.

        Args:
            buf: The bytes to parse.
            params: The parsing parameters.

        """
        try:
            tag, buf = Tag.parse(buf, params)
        except NotParseable as exc:
            return InvalidCommand(params, exc), buf[0:0]
        else:
            params = params.copy(tag=tag.value)
        cmd_parts: List[bytes] = []
        while True:
            try:
                _, buf = Space.parse(buf, params)
                atom, buf = Atom.parse(buf, params)
                cmd_parts.append(atom.value.upper())
            except NotParseable as exc:
                return InvalidCommand(params, exc), buf[0:0]
            command = b' '.join(cmd_parts)
            cmd_type = self.commands.get(command)
            if not cmd_type:
                return InvalidCommand(params, None, command), buf[0:0]
            elif not cmd_type.compound:
                break
        params = params.copy(command_name=command)
        try:
            return cmd_type.parse(buf, params)
        except NotParseable as exc:
            return InvalidCommand(params, exc, command, cmd_type), buf[0:0]




from __future__ import annotations

import asyncio
import json
import uuid
from argparse import Namespace
from asyncio import Task
from contextlib import closing, asynccontextmanager
from functools import partial
from typing import Any, Optional, Tuple, Mapping, AsyncIterator

from aioredis import create_redis, Redis  # type: ignore
from pysasl import AuthenticationCredentials

from pymap.bytes import BytesFormat
from pymap.config import BackendCapability, IMAPConfig
from pymap.context import connection_exit
from pymap.exceptions import InvalidAuth, IncompatibleData, MailboxConflict
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.session import LoginProtocol

from .cleanup import Cleanup, CleanupTask
from .filter import FilterSet
from .keys import DATA_VERSION, RedisKey, GlobalKeys, NamespaceKeys
from .mailbox import Message, MailboxSet
from ..session import BaseSession

try:
    from passlib.apps import ldap_context  # type: ignore
except ImportError:
    ldap_context = None

__all__ = ['RedisBackend', 'Config', 'Session']


class RedisBackend(BackendInterface):
    """Defines a backend that uses redis data structures for mailbox storage.

    """

    def __init__(self, login: LoginProtocol, config: Config) -> None:
        super().__init__()
        self._login = login
        self._config = config
        self._task = asyncio.create_task(self._run())

    @property
    def login(self) -> LoginProtocol:
        return self._login

    @property
    def config(self) -> Config:
        return self._config

    @property
    def task(self) -> Task:
        return self._task

    async def _run(self) -> None:
        config = self._config
        global_keys = config._global_keys
        connect_redis = partial(create_redis, config.address)
        await CleanupTask(connect_redis, global_keys).run_forever()

    @classmethod
    def add_subparser(cls, name: str, subparsers) -> None:
        parser = subparsers.add_parser(name, help='redis backend')
        parser.add_argument('address', nargs='?', default='redis://localhost',
                            help='the redis server address')
        parser.add_argument('--select', metavar='DB', type=int,
                            help='the redis database for mail data')
        parser.add_argument('--separator', metavar='CHAR', default='/',
                            help='the redis key segment separator')
        parser.add_argument('--prefix', metavar='VAL', default='',
                            help='the mail data key prefix')
        parser.add_argument('--users-prefix', metavar='VAL', default='',
                            help='the user lookup key prefix')
        parser.add_argument('--users-json', action='store_true',
                            help='the user lookup value contains JSON')

    @classmethod
    async def init(cls, args: Namespace) -> Tuple[RedisBackend, Config]:
        config = Config.from_args(args)
        return cls(Session.login, config), config


class Config(IMAPConfig):
    """The config implementation for the redis backend.

    Args:
        args: The command-line arguments.
        address: The redis server address.
        select: The redis database for mail data.
        separator: The redis key segment separator.
        prefix: The prefix for mail data keys.
        users_prefix: The user lookup key template.
        users_json: True if the user lookup value contains JSON.

    """

    def __init__(self, args: Namespace, *, address: str, select: Optional[int],
                 separator: bytes, prefix: bytes, users_prefix: bytes,
                 users_json: bool, **extra: Any) -> None:
        super().__init__(args, **extra)
        self._address = address
        self._select = select
        self._separator = separator
        self._prefix = prefix
        self._users_prefix = users_prefix
        self._users_json = users_json

    @property
    def backend_capability(self) -> BackendCapability:
        return BackendCapability(idle=True, object_id=True, multi_append=True)

    @property
    def address(self) -> str:
        """The redis server address. Defaults to a connection to localhost.

        See Also:
            :func:`aioredis.create_connection`

        """
        return self._address

    @property
    def select(self) -> Optional[int]:
        """The redis database for mail data. If given, the `SELECT`_ command is
        called after successful user lookup.

        .. _SELECT: https://redis.io/commands/select

        """
        return self._select

    @property
    def separator(self) -> bytes:
        """The bytestring used to separate segments of composite redis keys."""
        return self._separator

    @property
    def prefix(self) -> bytes:
        """The prefix for mail data keys. This prefix does not apply to
        :attr:`.users_key`.

        """
        return self._prefix

    @property
    def users_prefix(self) -> bytes:
        """The prefix for user lookup keys."""
        return self._users_prefix

    @property
    def users_json(self) -> bool:
        """True if the value from the user lookup key contains a JSON object
        with a ``"password"`` attribute, instead of a redis hash with a
        ``password`` key.

        See Also:
            `redis hashes
            <https://redis.io/topics/data-types-intro#redis-hashes>`_

        """
        return self._users_json

    @property
    def _joiner(self) -> BytesFormat:
        return BytesFormat(self.separator)

    @property
    def _users_root(self) -> RedisKey:
        return RedisKey(self._joiner, [self.users_prefix], {})

    @property
    def _global_keys(self) -> GlobalKeys:
        key = RedisKey(self._joiner, [self.prefix], {})
        return GlobalKeys(key)

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        return {'address': args.address,
                'select': args.select,
                'separator': args.separator.encode('utf-8'),
                'prefix': args.prefix.encode('utf-8'),
                'users_prefix': args.users_prefix.encode('utf-8'),
                'users_json': args.users_json}


class Session(BaseSession[Message]):
    """The session implementation for the redis backend."""

    resource = __name__

    _namespace_key = 'pymap:namespace'
    _version_key = 'pymap:dataversion'

    def __init__(self, redis: Redis, owner: str, config: Config,
                 mailbox_set: MailboxSet, filter_set: FilterSet) -> None:
        super().__init__(owner)
        self._redis = redis
        self._config = config
        self._mailbox_set = mailbox_set
        self._filter_set = filter_set

    @property
    def config(self) -> IMAPConfig:
        return self._config

    @property
    def mailbox_set(self) -> MailboxSet:
        return self._mailbox_set

    @property
    def filter_set(self) -> FilterSet:
        return self._filter_set

    @classmethod
    async def _connect_redis(cls, address: str) -> Redis:
        redis = await create_redis(address)
        stack = connection_exit.get()
        stack.enter_context(closing(redis))
        return redis

    @classmethod
    @asynccontextmanager
    async def login(cls, credentials: AuthenticationCredentials,
                    config: Config) -> AsyncIterator[Session]:
        """Checks the given credentials for a valid login and returns a new
        session.

        """
        redis = await cls._connect_redis(config.address)
        user = await cls._check_user(redis, config, credentials)
        if config.select is not None:
            await redis.select(config.select)
        global_keys = config._global_keys
        namespace = await cls._get_namespace(redis, global_keys, user)
        ns_keys = NamespaceKeys(global_keys, namespace)
        cleanup = Cleanup(global_keys)
        mailbox_set = MailboxSet(redis, ns_keys, cleanup)
        try:
            await mailbox_set.add_mailbox('INBOX')
        except MailboxConflict:
            pass
        filter_set = FilterSet(redis, ns_keys)
        yield cls(redis, credentials.identity, config, mailbox_set, filter_set)

    @classmethod
    async def _check_user(cls, redis: Redis, config: Config,
                          credentials: AuthenticationCredentials) -> str:
        user = credentials.authcid
        password = await cls._get_password(redis, config, user)
        if user != credentials.identity:
            raise InvalidAuth()
        elif ldap_context is None or not credentials.has_secret:
            if not credentials.check_secret(password):
                raise InvalidAuth()
        elif not ldap_context.verify(credentials.secret, password):
            raise InvalidAuth()
        return user

    @classmethod
    async def _get_password(cls, redis: Redis, config: Config,
                            user: str) -> str:
        user_key = config._users_root.end(user.encode('utf-8'))
        if config.users_json:
            json_data = await redis.get(user_key)
            if json_data is None:
                raise InvalidAuth()
            try:
                json_obj = json.loads(json_data)
                return json_obj['password']
            except Exception as exc:
                raise InvalidAuth() from exc
        else:
            password, identity = await redis.hmget(
                user_key, b'password', b'identity')
            if password is None:
                raise InvalidAuth()
            return password.decode('utf-8')

    @classmethod
    async def _get_namespace(cls, redis: Redis, global_keys: GlobalKeys,
                             user: str) -> bytes:
        user_key = user.encode('utf-8')
        new_namespace = uuid.uuid4().hex.encode('ascii')
        ns_val = b'%d/%b' % (DATA_VERSION, new_namespace)
        multi = redis.multi_exec()
        multi.hsetnx(global_keys.namespaces, user_key, ns_val)
        multi.hget(global_keys.namespaces, user_key)
        _, ns_val = await multi.execute()
        version, namespace = ns_val.split(b'/', 1)
        if int(version) != DATA_VERSION:
            raise IncompatibleData()
        return namespace




from __future__ import annotations

from contextlib import AsyncExitStack

from pymap.context import connection_exit

__all__ = ['exit_context']


def exit_context(func):
    async def deco(*args, **kwargs):
        async with AsyncExitStack() as stack:
            connection_exit.set(stack)
            return await func(*args, **kwargs)
    return deco




from __future__ import annotations

import uuid
from typing import Optional, Tuple, Sequence, List

from aioredis import Redis, MultiExecError  # type: ignore

from ._util import unwatch_pipe, watch_pipe, check_errors
from .keys import NamespaceKeys, FilterKeys
from pymap.filter import EntryPointFilterSet

__all__ = ['FilterSet']


class FilterSet(EntryPointFilterSet[bytes]):

    _active_name = b''

    def __init__(self, redis: Redis, ns_keys: NamespaceKeys) -> None:
        super().__init__('sieve', bytes)
        self._redis = redis
        self._keys = FilterKeys(ns_keys)

    @classmethod
    def _get_random_key(cls) -> bytes:
        return uuid.uuid4().hex.encode('ascii')

    async def put(self, name: str, value: bytes) -> None:
        redis = self._redis
        keys = self._keys
        name_bytes = name.encode('utf-8')
        key = self._get_random_key()
        pipe = unwatch_pipe(redis)
        pipe.hset(keys.data, key, value)
        pipe.hset(keys.names, name_bytes, key)
        await pipe.execute()

    async def delete(self, name: str) -> None:
        redis = self._redis
        keys = self._keys
        name_bytes = name.encode('utf-8')
        while True:
            pipe = watch_pipe(redis, keys.names)
            pipe.hmget(keys.names, name_bytes, self._active_name)
            _, _, (name_key, active_key) = await pipe.execute()
            if name_key is None:
                raise KeyError(name)
            elif name_key == active_key:
                raise ValueError(name)
            multi = redis.multi_exec()
            multi.hdel(keys.data, name_key)
            multi.hdel(keys.names, name_bytes)
            try:
                await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                break

    async def rename(self, before_name: str, after_name: str) -> None:
        redis = self._redis
        keys = self._keys
        before_name_bytes = before_name.encode('utf-8')
        after_name_bytes = after_name.encode('utf-8')
        while True:
            pipe = watch_pipe(redis, keys.names)
            pipe.hmget(keys.names, before_name_bytes, after_name_bytes)
            _, _, (before_name_key, after_name_key) = await pipe.execute()
            if before_name_key is None:
                raise KeyError(before_name)
            elif after_name_key is not None:
                raise KeyError(after_name)
            multi = redis.multi_exec()
            multi.hdel(keys.names, before_name_bytes)
            multi.hset(keys.names, after_name_bytes, before_name_key)
            try:
                await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                break

    async def clear_active(self) -> None:
        redis = self._redis
        keys = self._keys
        pipe = unwatch_pipe(redis)
        pipe.hdel(keys.names, self._active_name)
        await pipe.execute()

    async def set_active(self, name: str) -> None:
        redis = self._redis
        keys = self._keys
        name_bytes = name.encode('utf-8')
        while True:
            pipe = watch_pipe(redis, keys.names)
            pipe.hget(keys.names, name_bytes)
            _, _, name_key = await pipe.execute()
            if name_key is None:
                raise KeyError(name)
            multi = redis.multi_exec()
            multi.hset(keys.names, self._active_name, name_key)
            try:
                await multi.execute()
            except MultiExecError:
                if await check_errors(multi):
                    raise
            else:
                break

    async def get(self, name: str) -> bytes:
        redis = self._redis
        keys = self._keys
        name_bytes = name.encode('utf-8')
        pipe = unwatch_pipe(redis)
        pipe.hget(keys.names, name_bytes)
        _, name_key = await pipe.execute()
        if name_key is None:
            raise KeyError(name)
        value = await redis.hget(keys.data, name_key)
        if value is None:
            raise KeyError(name)
        return value

    async def get_active(self) -> Optional[bytes]:
        redis = self._redis
        keys = self._keys
        pipe = unwatch_pipe(redis)
        pipe.hget(keys.names, self._active_name)
        _, active_key = await pipe.execute()
        if active_key is None:
            return None
        value = await redis.hget(keys.data, active_key)
        if value is None:
            return None
        return value

    async def get_all(self) -> Tuple[Optional[str], Sequence[str]]:
        redis = self._redis
        keys = self._keys
        pipe = unwatch_pipe(redis)
        pipe.hgetall(keys.names)
        _, sieve_names = await pipe.execute()
        active_key: Optional[bytes] = sieve_names.get(self._active_name)
        active_name: Optional[str] = None
        names: List[str] = []
        for name, key in sieve_names.items():
            if name != self._active_name:
                name_str = name.decode('utf-8')
                names.append(name_str)
                if key == active_key:
                    active_name = name_str
        return active_name, names




from __future__ import annotations

from datetime import datetime
from typing import Any, Optional, NamedTuple, FrozenSet

from ..parsing.specials import Flag, ExtensionOptions, ObjectId

__all__ = ['AppendMessage', 'PreparedMessage']


class AppendMessage(NamedTuple):
    """A single message from the APPEND command.

    Args:
        literal: The message literal.
        when: The internal timestamp to assign to the message.
        flag_set: The flags to assign to the message.
        options: The extension options in use for the message.

    """

    literal: bytes
    when: Optional[datetime]
    flag_set: FrozenSet[Flag]
    options: Optional[ExtensionOptions] = None


class PreparedMessage(NamedTuple):
    """A message that has been prepared for appending to a mailbox.

    Args:
        when: The internal timestamp to assign to the message.
        flag_set: The flags to assign to the message.
        email_id: An email object ID to assign to the message.
        thread_id: A thread object ID to assign to the message.
        options: The extension options in use for the message.
        ref: A strong reference to an object that will be held until the append
            operation succeeds or fails, for use with :func:`weakref.finalize`.

    """

    when: Optional[datetime]
    flag_set: FrozenSet[Flag]
    email_id: ObjectId
    thread_id: ObjectId
    options: Optional[ExtensionOptions] = None
    ref: Any = None




from __future__ import annotations

from abc import abstractmethod, ABCMeta
from typing import TypeVar, Generic, Any, Type, Optional, Tuple, Sequence
from typing_extensions import Final

__all__ = ['ParsingExpectedT', 'ParsingState', 'ParsingInterrupt',
           'ParsingExpectation']

#: Type variable for expected data during parsing.
ParsingExpectedT = TypeVar('ParsingExpectedT')


class ParsingState:
    """Contains mutable parsing state. As parsing advances, it may need
    external data to proceed. Because parsing is linear, these expectations
    are defined as a mapping of an key string to a sequence of objects.

    Warning:
        The expectation keys and object types are dynamic, to prevent cyclic
        dependencies in the parsing system.

    See Also:
        :class:`pymap.parsing.interrupts.ParsingInterrupt`

    Args:
        expectations: The expect

    """

    __slots__ = ['_expectations', '_iters']

    def __init__(self, **expectations: Sequence[Any]) \
            -> None:
        super().__init__()
        self._iters = {key: iter(val) for key, val in expectations.items()}

    def consume(self, key: str, cls: Type[ParsingExpectedT]) \
            -> Optional[ParsingExpectedT]:
        """Consume and return a piece of data, if available.

        Args:
            key: The expectation key.

        """
        try:
            return next(self._iters[key])
        except (KeyError, StopIteration):
            return None


class ParsingInterrupt(Exception):
    """Exception thrown to interrupt the IMAP parsing process, because an
    expectation was not met. The caller should update provide an updated
    :class:`ParsingState` object that meets the expectation.

    Args:
        expected: The required data expectation.

    """

    __slots__ = ['expected']

    def __init__(self, expected: ParsingExpectation) -> None:
        super().__init__()
        self.expected: Final = expected


class ParsingExpectation(Generic[ParsingExpectedT], metaclass=ABCMeta):
    """Base class for parsing expectations that may require additional data
    before parsing can advance.

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def consume_args(self) -> Tuple[str, Type[ParsingExpectedT]]:
        """Tuple of the *key* and *cls* arguments to pass to
        :meth:`~ParsingState.consume`.

        """
        ...

    def expect(self, state: ParsingState) -> ParsingExpectedT:
        """Indicates that the buffer has been successfully parsed so far, but
        additional data is expected from the parsing state.

        Args:
            state: The parsing state object.

        Raises:
            ParsingInterrupt: The expected data was not available.

        """
        next_val = state.consume(*self.consume_args)
        if next_val is None:
            raise ParsingInterrupt(self)
        else:
            return next_val




from __future__ import annotations

from typing import Tuple, Type
from typing_extensions import Final

from . import ParsingExpectation
from ..message import AppendMessage, PreparedMessage

__all__ = ['ExpectPreparedMessage']


class ExpectPreparedMessage(ParsingExpectation[PreparedMessage]):
    """Indicates that the APPEND command has parsed a message that must be
    prepared before processing can proceed.

    Args:
        name: The destination mailbox name.
        append_msg: The appended message data.

    """

    __slots__ = ['mailbox', 'message']

    def __init__(self, mailbox: str, message: AppendMessage) -> None:
        super().__init__()
        self.mailbox: Final = mailbox
        self.message: Final = message

    @property
    def consume_args(self) -> Tuple[str, Type[PreparedMessage]]:
        return 'prepared_messages', PreparedMessage




from __future__ import annotations

import asyncio
import logging
from contextlib import closing
from typing import ClassVar, Callable, Awaitable, NoReturn
from typing_extensions import Final

from aioredis import Redis, ConnectionClosedError  # type: ignore

from ._util import unwatch_pipe
from .keys import RedisKey, GlobalKeys, CleanupKeys, NamespaceKeys, \
    ContentKeys, MailboxKeys, MessageKeys

__all__ = ['Cleanup', 'CleanupTask', 'CleanupThread']

_log = logging.getLogger(__name__)


class Cleanup:
    """Defines the logic for adding key groups to be cleaned up.

    Args:
        root: The root redis key.

    """

    content_expire = 3600

    def __init__(self, global_keys: GlobalKeys) -> None:
        super().__init__()
        keys = CleanupKeys(global_keys)
        self.keys: Final = keys
        self._order = (keys.messages, keys.mailboxes, keys.namespaces,
                       keys.contents, keys.roots)

    def add_namespace(self, pipe: Redis, keys: NamespaceKeys) -> None:
        """Add the namespace to be cleaned up.

        Args:
            pipe: Piped redis commands.
            keys: The namespace key group.

        """
        cleanup_val = keys.root.named['namespace']
        pipe.rpush(self.keys.namespaces, cleanup_val)

    def add_mailbox(self, pipe: Redis, keys: MailboxKeys) -> None:
        """Add the mailbox to be cleaned up.

        Args:
            pipe: Piped redis commands.
            keys: The mailbox key group.

        """
        namespace = keys.root.named['namespace']
        mailbox_id = keys.root.named['mailbox_id']
        cleanup_val = b'%b\x00%b' % (namespace, mailbox_id)
        pipe.rpush(self.keys.mailboxes, cleanup_val)

    def add_message(self, pipe: Redis, keys: MessageKeys) -> None:
        """Add the message to be cleaned up.

        Args:
            pipe: Piped redis commands.
            keys: The message key group.

        """
        namespace = keys.root.named['namespace']
        mailbox_id = keys.root.named['mailbox_id']
        msg_uid = keys.root.named['uid']
        cleanup_val = b'%b\x00%b\x00%b' \
            % (namespace, mailbox_id, msg_uid)
        pipe.rpush(self.keys.messages, cleanup_val)

    def add_content(self, pipe: Redis, keys: ContentKeys) -> None:
        """Add the content to be cleaned up.

        Args:
            pipe: Piped redis commands.
            keys: The content key group.

        """
        namespace = keys.root.named['namespace']
        email_id = keys.root.named['email_id']
        cleanup_val = b'%b\x00%b' % (namespace, email_id)
        pipe.rpush(self.keys.contents, cleanup_val)

    def add_root(self, pipe: Redis, root: RedisKey) -> None:
        """Add the content to be cleaned up.

        Args:
            pipe: Piped redis commands.
            root: The redis key prefix.

        """
        cleanup_val = root.wildcard
        pipe.rpush(self.keys.roots, cleanup_val)


class CleanupTask:
    """Maintains a :class:`CleanupThread` for the duration of the process
    lifetime, restarting on failure.

    Args:
        connect_redis: Supplies a connected redis object.
        root: The root redis key.

    """

    #: The delay between redis reconnect attempts, on connection failure.
    connection_delay: ClassVar[float] = 5.0

    def __init__(self, connect_redis: Callable[[], Awaitable[Redis]],
                 global_keys: GlobalKeys) -> None:
        super().__init__()
        self._connect_redis = connect_redis
        self._global_keys = global_keys

    async def run_forever(self) -> NoReturn:
        """Run the cleanup loop indefinitely."""
        while True:
            try:
                with closing(await self._connect_redis()) as redis:
                    await CleanupThread(redis, self._global_keys).run()
            except (ConnectionClosedError, OSError):
                _log.warning('Redis connection failure', exc_info=True)
            await asyncio.sleep(self.connection_delay)


class CleanupThread:
    """Defines the logic for monitoring and executing cleanup of various
    entities.

    Args:
        redis: The redis connection object.
        global_keys: The global keys group.

    """

    def __init__(self, redis: Redis, global_keys: GlobalKeys) -> None:
        super().__init__()
        self._redis = redis
        self._cleanup = Cleanup(global_keys)
        self._global_keys = global_keys

    async def run(self) -> NoReturn:
        """Run the cleanup loop indefinitely.

        Raises:
            :class:`~aioredis.ConnectionClosedError`: The connection to redis
                was interrupted.

        """
        redis = self._redis
        cleanup = self._cleanup
        while True:
            cleanup_key, cleanup_val = await redis.blpop(
                *cleanup._order, timeout=0)
            try:
                await asyncio.shield(self._run_one(cleanup_key, cleanup_val))
            except Exception:
                _log.warning('Cleanup failed: key=%s val=%s',
                             cleanup_key, cleanup_val, exc_info=True)
                raise

    async def _run_one(self, cleanup_key: bytes, cleanup_val: bytes) -> None:
        cleanup = self._cleanup
        if cleanup_key == cleanup.keys.namespaces:
            namespace = cleanup_val
            await self._run_namespace(namespace)
        elif cleanup_key == cleanup.keys.mailboxes:
            namespace, mailbox_id = cleanup_val.split(b'\x00', 1)
            await self._run_mailbox(namespace, mailbox_id)
        elif cleanup_key == cleanup.keys.messages:
            namespace, mailbox_id, msg_uid = cleanup_val.split(b'\x00', 2)
            await self._run_message(namespace, mailbox_id, msg_uid)
        elif cleanup_key == cleanup.keys.contents:
            namespace, email_id = cleanup_val.split(b'\x00', 1)
            await self._run_content(namespace, email_id)
        elif cleanup_key == cleanup.keys.roots:
            wildcard = cleanup_val
            await self._run_root(wildcard)

    async def _run_namespace(self, namespace: bytes) -> None:
        redis = self._redis
        cleanup = self._cleanup
        ns_keys = NamespaceKeys(self._global_keys, namespace)
        pipe = unwatch_pipe(redis)
        pipe.hvals(ns_keys.mailboxes)
        _, mailbox_ids = await pipe.execute()
        multi = redis.multi_exec()
        multi.unlink(*ns_keys.keys)
        for mailbox_id in mailbox_ids:
            mbx_keys = MailboxKeys(ns_keys, mailbox_id)
            cleanup.add_mailbox(multi, mbx_keys)
        cleanup.add_root(multi, ns_keys.root)
        await multi.execute()

    async def _run_mailbox(self, namespace: bytes, mailbox_id: bytes) -> None:
        redis = self._redis
        cleanup = self._cleanup
        ns_keys = NamespaceKeys(self._global_keys, namespace)
        mbx_keys = MailboxKeys(ns_keys, mailbox_id)
        pipe = unwatch_pipe(redis)
        pipe.smembers(mbx_keys.uids)
        _, msg_uids = await pipe.execute()
        multi = redis.multi_exec()
        multi.unlink(*mbx_keys.keys)
        for msg_uid in msg_uids:
            msg_keys = MessageKeys(mbx_keys, msg_uid)
            cleanup.add_message(multi, msg_keys)
        cleanup.add_root(multi, mbx_keys.root)
        await multi.execute()

    async def _run_message(self, namespace: bytes, mailbox_id: bytes,
                           msg_uid: bytes) -> None:
        redis = self._redis
        cleanup = self._cleanup
        ns_keys = NamespaceKeys(self._global_keys, namespace)
        mbx_keys = MailboxKeys(ns_keys, mailbox_id)
        msg_keys = MessageKeys(mbx_keys, msg_uid)
        pipe = unwatch_pipe(redis)
        pipe.hget(msg_keys.immutable, b'emailid')
        _, email_id = await pipe.execute()
        multi = redis.multi_exec()
        multi.unlink(*msg_keys.keys)
        if email_id is not None:
            ct_keys = ContentKeys(ns_keys, email_id)
            cleanup.add_content(multi, ct_keys)
        cleanup.add_root(multi, msg_keys.root)
        await multi.execute()

    async def _run_content(self, namespace: bytes, email_id: bytes) -> None:
        redis = self._redis
        cleanup = self._cleanup
        ns_keys = NamespaceKeys(self._global_keys, namespace)
        ct_keys = ContentKeys(ns_keys, email_id)
        pipe = unwatch_pipe(redis)
        pipe.ttl(ct_keys.data)
        pipe.hincrby(ns_keys.content_refs, email_id, -1)
        _, ttl, refs = await pipe.execute()
        if ttl < 0 and int(refs or 0) <= 0:
            await redis.expire(ct_keys.data, cleanup.content_expire)

    async def _run_root(self, wildcard: bytes) -> None:
        redis = self._redis
        cur = b'0'
        await redis.unwatch()
        while cur:
            cur, keys = await redis.scan(cur, match=wildcard)
            if keys:
                await redis.unlink(*keys)
